/*
* Chip8 emulator
* 
* Copyright (C) 2022 rick G. <rick@gnous.eu>
* 
* This program is free software: you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation, either version 3 of the License, or (at your option) any later
* version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with
* this program. If not, see <https://www.gnu.org/licenses/>.
*/

mod machine;
extern crate sdl2;

use sdl2::pixels::Color;
use sdl2::event::Event;
use sdl2::keyboard::Keycode;
use sdl2::rect::Point;

use machine::HEIGHT;
use machine::WIDTH;
use machine::Machine;

use std::path::Path;
use std::io;
use std::env;

fn main() {
    let name_file = &env::args().collect::<Vec<String>>()[1];
    //read_file(Path::new("15PUZZLE"));
    //return;
    let mut t: Machine = Machine::new();
    t.init();
    match t.load_rom(Path::new(name_file)) {
        Ok(()) => println!("La ROM a correctement été chargée."),
        Err(e) => match e.kind() {
            std::io::ErrorKind::NotFound => panic!("Le fichier n'a pas été trouvé."),
            _ => panic!("Erreur inconnue : {}", e),
        }
    }

    let sdl_context = sdl2::init().unwrap();
    let video_subsystem = sdl_context.video().unwrap();
    let window = video_subsystem.window("Chip8 Emulator", WIDTH as u32, HEIGHT as u32)
    //let window = video_subsystem.window("Chip8 Emulator", 100 + 10 * WIDTH as u32, 10 * HEIGHT as u32)
        .position_centered()
        .build()
        .unwrap();

    let mut canvas = window.into_canvas().build().unwrap();
    canvas.set_draw_color(Color::RGB(0, 0, 0));
    canvas.clear();
    canvas.present();

    let mut event_pump = sdl_context.event_pump().unwrap();

    'running: loop {
        t.cycle(&mut event_pump);
        
        for i in 0..t.display_buffer.len() {
            for j in 0..t.display_buffer[i].len() {
                if t.display_buffer[i][j] == 1 {
                    canvas.set_draw_color(Color::RGB(255, 255, 255));
                    canvas.draw_point(Point::new(j as i32, i as i32)).unwrap();
                } else {
                    canvas.set_draw_color(Color::RGB(0, 0, 0));
                    canvas.draw_point(Point::new(j as i32, i as i32)).unwrap();
                }
            }
        }

        for event in event_pump.poll_iter() {
            match event {
                Event::KeyDown { keycode: Some(Keycode::Escape), ..} => {
                    break 'running
                },
                Event::KeyDown { keycode: Some(Keycode::Num0), .. } => {
                    t.key_inputs[0] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num1), .. } => {
                    t.key_inputs[1] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num2), .. } => {
                    t.key_inputs[2] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num3), .. } => {
                    t.key_inputs[3] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num4), .. } => {
                    t.key_inputs[4] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num5), .. } => {
                    t.key_inputs[5] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num6), .. } => {
                    t.key_inputs[6] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num7), .. } => {
                    t.key_inputs[7] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num8), .. } => {
                    t.key_inputs[8] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::Num9), .. } => {
                    t.key_inputs[9] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::A), .. } => {
                    t.key_inputs[0xa] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::B), .. } => {
                    t.key_inputs[0xb] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::C), .. } => {
                    t.key_inputs[0xc] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::D), .. } => {
                    t.key_inputs[0xd] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::E), .. } => {
                    t.key_inputs[0xe] = 1;
                },
                Event::KeyDown { keycode: Some(Keycode::F), .. } => {
                    t.key_inputs[0xf] = 1;
                },
                Event::KeyUp { keycode: Some(Keycode::Num0), .. } => {
                    t.key_inputs[0] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num1), .. } => {
                    t.key_inputs[1] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num2), .. } => {
                    t.key_inputs[2] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num3), .. } => {
                    t.key_inputs[3] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num4), .. } => {
                    t.key_inputs[4] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num5), .. } => {
                    t.key_inputs[5] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num6), .. } => {
                    t.key_inputs[6] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num7), .. } => {
                    t.key_inputs[7] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num8), .. } => {
                    t.key_inputs[8] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::Num9), .. } => {
                    t.key_inputs[9] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::A), .. } => {
                    t.key_inputs[0xa] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::B), .. } => {
                    t.key_inputs[0xb] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::C), .. } => {
                    t.key_inputs[0xc] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::D), .. } => {
                    t.key_inputs[0xd] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::E), .. } => {
                    t.key_inputs[0xe] = 0;
                },
                Event::KeyUp { keycode: Some(Keycode::F), .. } => {
                    t.key_inputs[0xf] = 0;
                },
                _ => {}
            }
        }

        canvas.present();
    }
}

use std::fs::File;
use std::io::Read;
#[allow(dead_code, unused_variables)]
fn read_file(p: &Path) -> io::Result<()> {
    let mut buffer = Vec::new();
    let mut f = File::open(p)?;
    let mut mem = [0; 4096];
    f.read_to_end(&mut buffer)?;

    for i in 0..buffer.len() {
        mem[i] = buffer[i];
    }

    println!("{:?}", mem);

    let mut i = 0;
    while i < mem.len() {
        let opcode = ((mem[i] as u16) << 8) 
            + (mem[i + 1] as u16);

        let vx: usize = ((opcode & 0x0f00) >> 8) as usize;
        let vy: usize = ((opcode & 0x00f0) >> 4) as usize;
        let k: u8 = (opcode & 0x00ff) as u8;

        i += 2;

        match (opcode & 0xf000) >> 12 {
            0 => match opcode {
                0x00E0 => println!("CLS"),
                0x00EE => println!("RET"),
                _ => println!("SYS"),
            }
            1 => println!("JP {}", opcode & 0x0fff),
            2 => println!("CALL {}", opcode & 0x0fff),
            3 => println!("3SE {} {}", vx, k),
            4 => println!("4SNE {} {}", vx, k),
            5 => println!("5SE {} {}", vx, vy),
            6 => println!("6LD {} {}", vx, k),
            7 => println!("7ADD {} {}", vx, k),
            8 => {
                match opcode & 0x000f {
                    0 => println!("8LD {} {}", vx, vy),
                    1 => println!("8OR {} {}", vx, vy),
                    2 => println!("8AND {} {}", vx, vy),
                    3 => println!("8XOR {} {}", vx, vy),
                    4 => println!("8ADD {} {}", vx, vy),
                    5 => println!("8SUB {} {}", vx, vy),
                    6 => println!("8SHR {} {}", vx, vy),
                    7 => println!("8SUBN {} {}", vx, vy),
                    0xE => println!("8SHL {} {}", vx, vy),
                    _ => println!("error 8")
                }
            }
            9 => println!("9SNE {} {}", vx, vy),
            0xA => println!("AADD I {}", opcode & 0x0fff),
            0xB => println!("BJP 0 {}", opcode & 0x0fff),
            0xC => println!("RND {} {}", vx, k),
            0xD => println!("DRW {} {} {}", vx, vy, opcode & 0x000f),
            0xE => {
                let x = (opcode & 0x0f00) >> 8;
                match opcode & 0x00ff {
                    0x9E => println!("SKP {}", vx),
                    0xA1 => println!("SKNP {}", vx),
                    _ => println!("error E")
                }
            }
            0xF => {
                let x = (opcode & 0x0f00) >> 8;
                match opcode & 0x00ff {
                    0x07 => println!("FLD {} DT", vx),
                    0x0A => println!("FLD {} K", vx),
                    0x15 => println!("FLD DT {}", vx),
                    0x18 => println!("FLD ST {}", vx),
                    0x1E => println!("FADD I {}", vx),
                    0x29 => println!("FLD F {}", vx),
                    0x33 => println!("FLD B {}", vx),
                    0x55 => println!("FLD [I] {}", vx),
                    0x65 => println!("FLD {} [I]", vx),
                    _ => println!("error F")
                }
            }
            _ => println!("error")
        }
    }

    Ok(())
}
